from pydantic import BaseModel, Field
import typing as t

from clonf.annotations import ClonfAnnotation
from clonf.extractor import extract_cli_info


def test_extract_clonf_annotation() -> None:
    class SimpleModel(BaseModel):
        debug: t.Annotated[bool, Field(), ClonfAnnotation()]

    cli_info = extract_cli_info(SimpleModel)
    assert len(cli_info) == 1
    for ann in cli_info:
        assert isinstance(ann, ClonfAnnotation)



from __future__ import annotations
from dataclasses import dataclass, field
import typing as t
from pydantic.fields import FieldInfo


@dataclass
class ClonfAnnotation:
    name: str | ellipsis = field(default=...)
    description: str | None = field(default=None)
    _type: t.Any | None = field(default=None, init=False)
    _field_info: FieldInfo | None = field(default=None, init=False)


@dataclass
class CliArgument(ClonfAnnotation):
    default: t.Any = field(default=...)


@dataclass
class CliOption(ClonfAnnotation):
    default: t.Any = field(default=...)
    prefix: str = field(default="--")



from __future__ import annotations

from pydantic import BaseModel
import functools


from ...annotations import ClonfAnnotation, CliArgument, CliOption
from ...extractor import extract_cli_info
import typing as t
import datetime
import pathlib
import uuid
from .types import _CliFunc, _TReturn, _WrappedFunc

try:
    import click
except ImportError:  # pragma: no cover
    raise ImportError("clonf integration with click requires click to be installed")


def _extract_cli_info_click(model: type[BaseModel]) -> list[ClonfAnnotation]:
    cli_infos = extract_cli_info(model)

    for cli_info in cli_infos:
        if cli_info._type is None or cli_info._field_info is None:  # pragma: no cover
            continue

        for meta in cli_info._field_info.metadata:
            if isinstance(meta, click.ParamType):
                cli_info._type = meta
                break

        if isinstance(cli_info._type, click.ParamType):
            continue
        type_origin = t.get_origin(cli_info._type)
        if type_origin is t.Literal:
            literal_values = t.get_args(cli_info._type)
            cli_info._type = click.Choice(*literal_values)
        elif type_origin is not None:
            raise TypeError(f"Don't know how to handle type {type_origin}")
        elif issubclass(cli_info._type, (datetime.datetime, datetime.date)):
            cli_info._type = click.DateTime()
        elif issubclass(cli_info._type, pathlib.Path):
            cli_info._type = click.Path(path_type=pathlib.Path)
        elif issubclass(cli_info._type, uuid.UUID):
            cli_info._type = click.UUID
        elif issubclass(cli_info._type, bool):
            cli_info._type = click.BOOL
        elif issubclass(cli_info._type, int) or issubclass(cli_info._type, float):
            range_kwargs = {
                "min": cli_info._field_info._attributes_set.get("gt", None)
                or cli_info._field_info._attributes_set.get("ge", None),
                "max": cli_info._field_info._attributes_set.get("lt", None)
                or cli_info._field_info._attributes_set.get("le", None),
                "min_open": cli_info._field_info._attributes_set.get("ge", None)
                is not None,
                "max_open": cli_info._field_info._attributes_set.get("le", None)
                is not None,
            }

            if issubclass(cli_info._type, int):
                cli_info._type = click.IntRange(**range_kwargs)  # type: ignore[arg-type]
            elif issubclass(cli_info._type, float):
                cli_info._type = click.FloatRange(**range_kwargs)  # type: ignore[arg-type]

    return cli_infos


def clonf_click(
    func: _CliFunc[_TReturn],
) -> _WrappedFunc[_TReturn]:
    # TODO: replace with inspect.get_annotations when deprecating PY39
    func_annotations: dict[str, t.Any] | None = getattr(func, "__annotations__", None)
    if func_annotations is None:  # pragma: no cover
        raise ValueError("Could not extract annotations from input func")

    cli_infos: dict[str, list[ClonfAnnotation]] = {}
    for arg, ann in func_annotations.items():
        if arg == "return":
            continue
        assert issubclass(ann, BaseModel), (
            f"Input type {ann} is not a subclass of BaseModel"
        )

        cli_infos[arg] = _extract_cli_info_click(ann)

    @functools.wraps(func)
    @click.pass_context
    def wrapper(ctx: click.Context, /, **kwargs: t.Any) -> _TReturn:
        new_kwargs: dict[str, BaseModel] = {}

        for arg, ann in func_annotations.items():
            if arg == "return":
                continue
            cli_info_names: set[str] = {
                cli_info.name
                for cli_info in cli_infos[arg]
                if cli_info.name is not Ellipsis
            }
            selected_kwargs = {k: v for k, v in kwargs.items() if k in cli_info_names}
            to_remove: list[str] = []
            for kwarg, v in selected_kwargs.items():
                param: click.Parameter | None = None
                for s_param in ctx.command.params:
                    if s_param.name == kwarg:
                        param = s_param
                        break
                if param is None:  # pragma: no cover
                    raise ValueError(f"Could not find parameter for input: {kwarg}")
                default = param.get_default(ctx)
                if default == v:
                    to_remove.append(kwarg)

            for kwarg in to_remove:
                del selected_kwargs[kwarg]

            new_kwargs[arg] = ann(**selected_kwargs)

        result = func(**new_kwargs)

        return result

    for sub_cli_infos in cli_infos.values():
        for cli_info in sub_cli_infos:
            if isinstance(cli_info, CliArgument):
                assert isinstance(cli_info.name, str), (
                    "CliArgument has no resolved name"
                )
                click.argument(
                    cli_info.name,
                    default=None if cli_info.default is Ellipsis else cli_info.default,
                    type=cli_info._type,
                )(wrapper)
            elif isinstance(cli_info, CliOption):
                assert isinstance(cli_info.name, str), "CliOption has no resolved name"
                click.option(
                    f"{cli_info.prefix}{cli_info.name}",
                    default=None if cli_info.default is Ellipsis else cli_info.default,
                    type=cli_info._type,
                    help=cli_info.description,
                )(wrapper)

    return wrapper



import typing_extensions as te
import typing as t

_TReturn = t.TypeVar("_TReturn", covariant=True)


_CliFunc: te.TypeAlias = t.Callable[..., _TReturn]


class _WrappedFunc(t.Protocol, t.Generic[_TReturn]):
    def __call__(self, **kwargs: t.Any) -> _TReturn: ...


class TClonfClick(t.Protocol):
    def __call__(self, func: _CliFunc[_TReturn]) -> _WrappedFunc[_TReturn]: ...



from .annotations import CliArgument, CliOption
import typing as t
import importlib

if t.TYPE_CHECKING:
    from clonf.integrations.click.types import TClonfClick

    clonf_click: TClonfClick


def __getattr__(name: str) -> t.Any:
    if name == "clonf_click":
        return getattr(importlib.import_module("clonf.integrations.click"), name)

    raise AttributeError(  # pragma: no cover
        f"module {__name__!r} has no attribute {name!r}"
    )


__all__ = ["CliArgument", "CliOption", "clonf_click"]



import pathlib
import click
from click.testing import CliRunner
from clonf import clonf_click
from pydantic import BaseModel, Field
import typing as t
from clonf import CliArgument, CliOption
import uuid
import datetime
from pydantic_settings import BaseSettings, SettingsConfigDict
import os
import pytest

runner = CliRunner()


def test_simple_click_app() -> None:
    class SimpleConfig(BaseModel):
        name: t.Annotated[str, CliArgument()] = "World"
        greeting_word: t.Annotated[str, CliOption()] = "Hello"

    @click.command()
    @clonf_click
    def hello(config: SimpleConfig) -> None:
        click.echo(f"{config.greeting_word} {config.name}!")

    result = runner.invoke(hello, ["Guido"], catch_exceptions=False)
    assert result.exit_code == 0

    assert result.output == "Hello Guido!\n"

    result = runner.invoke(hello, catch_exceptions=False)
    assert result.exit_code == 0

    assert result.output == "Hello World!\n"


def test_click_int_range() -> None:
    class SimpleConfig(BaseModel):
        value: t.Annotated[int, Field(gt=1, lt=10), CliArgument()] = 5

    @click.command()
    @clonf_click
    def what_is_value(config: SimpleConfig) -> None:
        click.echo(f"The value is {config.value}")

    assert len(what_is_value.params) == 1
    assert isinstance(what_is_value.params[0].type, click.IntRange)


def test_click_float_range() -> None:
    class SimpleConfig(BaseModel):
        value: t.Annotated[float, Field(gt=1, lt=10), CliArgument()] = 5

    @click.command()
    @clonf_click
    def what_is_value(config: SimpleConfig) -> None:
        click.echo(f"The value is {config.value}")

    assert len(what_is_value.params) == 1
    assert isinstance(what_is_value.params[0].type, click.FloatRange)


def test_click_literal() -> None:
    class SimpleConfig(BaseModel):
        answer: t.Annotated[t.Literal["yes", "no"], CliArgument()] = "yes"

    @click.command()
    @clonf_click
    def what_is_answer(config: SimpleConfig) -> None:
        click.echo(f"The answer is {config.answer}")

    assert len(what_is_answer.params) == 1
    assert isinstance(what_is_answer.params[0].type, click.Choice)


def test_click_unsupported_type() -> None:
    class SimpleConfig(BaseModel):
        value: t.Annotated[t.Set[str], CliOption()]  # noqa: UP006

    with pytest.raises(TypeError):

        @click.command()
        @clonf_click
        def what_is_value(config: SimpleConfig) -> None:
            click.echo(f"The answer is {config.value}")

        runner.invoke(what_is_value)


def test_click_path() -> None:
    class SimpleConfig(BaseModel):
        value: t.Annotated[pathlib.Path, CliArgument()] = pathlib.Path()

    @click.command()
    @clonf_click
    def what_is_value(config: SimpleConfig) -> None:
        click.echo(f"The value is {config.value}")

    assert len(what_is_value.params) == 1
    assert isinstance(what_is_value.params[0].type, click.Path)


def test_click_path_exists() -> None:
    class SimpleConfig(BaseModel):
        value: t.Annotated[pathlib.Path, CliArgument(), click.Path(exists=True)] = (
            pathlib.Path()
        )

    @click.command()
    @clonf_click
    def what_is_value(config: SimpleConfig) -> None:
        click.echo(f"The value is {config.value}")

    assert len(what_is_value.params) == 1
    assert (
        isinstance(what_is_value.params[0].type, click.Path)
        and what_is_value.params[0].type.exists is True
    )


def test_click_uuid() -> None:
    class SimpleConfig(BaseModel):
        value: t.Annotated[uuid.UUID, CliArgument()] = uuid.uuid4()

    @click.command()
    @clonf_click
    def what_is_value(config: SimpleConfig) -> None:
        click.echo(f"The value is {config.value}")

    assert len(what_is_value.params) == 1
    assert what_is_value.params[0].type is click.UUID


def test_click_datetime() -> None:
    class SimpleConfig(BaseModel):
        value: t.Annotated[datetime.date, CliArgument()] = datetime.date(2025, 5, 4)

    @click.command()
    @clonf_click
    def what_is_value(config: SimpleConfig) -> None:
        click.echo(f"The value is {config.value}")

    assert len(what_is_value.params) == 1
    assert isinstance(what_is_value.params[0].type, click.DateTime)


def test_click_bool() -> None:
    class SimpleConfig(BaseModel):
        value: t.Annotated[bool, CliArgument()] = False

    @click.command()
    @clonf_click
    def what_is_value(config: SimpleConfig) -> None:
        click.echo(f"The value is {config.value}")

    assert len(what_is_value.params) == 1
    assert what_is_value.params[0].type is click.BOOL


def test_click_settings_sources_priority(tmp_path: pathlib.Path) -> None:
    env_file = tmp_path / ".env"
    env_file.write_text("TEST_VALUE=31")

    class Config(BaseSettings):
        value: t.Annotated[int, CliArgument()] = 42

        model_config = SettingsConfigDict(
            env_prefix="TEST_",
            case_sensitive=False,
            env_file=str(env_file),
        )

    @click.command
    @clonf_click
    def cli(config: Config) -> None:
        click.echo(f"The value is {config.value}")

    result = runner.invoke(cli, catch_exceptions=False)

    assert result.exit_code == 0

    assert result.output == "The value is 31\n"

    os.environ["TEST_VALUE"] = "31"
    result = runner.invoke(cli, catch_exceptions=False)

    assert result.exit_code == 0

    assert result.output == "The value is 31\n"


def test_click_multiple_models() -> None:
    class Config(BaseModel):
        value: t.Annotated[int, CliArgument()] = 42

    class Config2(BaseModel):
        value2: t.Annotated[int, CliArgument()] = 1337

    @click.command
    @clonf_click
    def cli(config: Config, config2: Config2) -> None:
        click.echo(f"{config.value} {config2.value2}")

    result = runner.invoke(cli, catch_exceptions=False)
    assert result.exit_code == 0
    assert result.output == "42 1337\n"



from .annotations import CliArgument, ClonfAnnotation, CliOption
from pydantic import BaseModel
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined


def _process_cli_argument(arg: CliArgument, field_info: FieldInfo) -> CliArgument:
    if arg.default is Ellipsis:
        new_default = field_info.default or field_info.default_factory
        if new_default is not PydanticUndefined:
            arg.default = new_default
    return arg


def _process_cli_option(option: CliOption, field_info: FieldInfo) -> CliOption:
    if option.default is Ellipsis:
        new_default = field_info.default or field_info.default_factory
        if new_default is not PydanticUndefined:
            option.default = new_default

    return option


def extract_cli_info(model: type[BaseModel]) -> list[ClonfAnnotation]:
    extracted: list[ClonfAnnotation] = []

    for field, field_info in model.model_fields.items():
        for meta in field_info.metadata:
            if not isinstance(meta, ClonfAnnotation):
                continue

            meta._field_info = field_info
            meta._type = field_info.annotation

            if meta.name is Ellipsis:
                meta.name = field_info.alias or field

            if meta.description is None:
                meta.description = field_info.description

            if isinstance(meta, CliArgument):
                meta = _process_cli_argument(meta, field_info)

            if isinstance(meta, CliOption):
                meta = _process_cli_option(meta, field_info)

            extracted.append(meta)

    return extracted
